import contextlib

import torch


def get_available_devices():
    devices = []

    if torch.cuda.is_available():
        devices.append("cuda")

    if hasattr(torch, "mlu") and torch.mlu.is_available():
        devices.append("mlu")

    return tuple(devices)


with contextlib.suppress(ImportError, ModuleNotFoundError):
    import torch_mlu  # noqa: F401



import pytest
import torch

import ninetoothed
from ninetoothed import Tensor, block_size
from ninetoothed.language import libdevice
from tests.utils import get_available_devices


def arrangement(input, exponent, output, BLOCK_SIZE=block_size()):
    return (
        input.tile((BLOCK_SIZE,)),
        exponent.tile((BLOCK_SIZE,)),
        output.tile((BLOCK_SIZE,)),
    )


def application(input, exponent, output):
    output = libdevice.pow(input, exponent)  # noqa: F841


def pow(input, exponent):
    output = torch.empty_like(input)

    pow_kernel = ninetoothed.make(
        arrangement, application, (Tensor(1), Tensor(1), Tensor(1))
    )

    pow_kernel(input, exponent, output)

    return output


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("dtype", (torch.float32,))
@pytest.mark.parametrize("size", (44925,))
def test(size, dtype, device):
    torch.manual_seed(0)

    input = torch.rand(size, dtype=dtype, device=device)
    exponent = torch.rand(size, dtype=dtype, device=device)

    output = pow(input, exponent)
    expected = torch.pow(input, exponent)

    assert torch.allclose(output, expected, equal_nan=True)



import ctypes
import functools
import subprocess

import pytest
import torch
import torch.nn.functional as F

import ninetoothed
import ninetoothed.generation
import tests.test_addmm as addmm
import tests.test_attention as attention
import tests.test_conv2d as conv2d
import tests.test_matmul as matmul
from ninetoothed import Tensor
from tests.utils import get_available_devices


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize(
    "dtype, ninetoothed_dtype", ((torch.bfloat16, ninetoothed.bfloat16),)
)
@pytest.mark.parametrize("size", (45327,))
def test_add(size, dtype, device, ninetoothed_dtype):
    torch.manual_seed(0)

    def _arrangement(input, other, output):
        def _arrange(tensor):
            return tensor.tile((256,))

        return _arrange(input), _arrange(other), _arrange(output)

    def _application(input, other, output):
        output = input + other  # noqa: F841

    tensors = tuple(Tensor(1, dtype=ninetoothed_dtype) for _ in range(3))
    caller = device
    kernel_name = "add"
    output_dir = ninetoothed.generation.CACHE_DIR

    launch_func = _generate_launch_func(
        _arrangement,
        _application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    shape = (size,)

    input = torch.randn(shape, dtype=dtype, device=device)
    other = torch.randn(shape, dtype=dtype, device=device)
    output = torch.empty_like(input)

    _run_launch_func(launch_func, input, other, output)

    expected = torch.add(input, other)

    assert torch.allclose(output, expected)


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize(
    "dtype, ninetoothed_dtype, atol", ((torch.float16, ninetoothed.float16, 0.075),)
)
@pytest.mark.parametrize("k", (512,))
@pytest.mark.parametrize("n", (512,))
@pytest.mark.parametrize("m", (512,))
def test_addmm(m, n, k, dtype, device, ninetoothed_dtype, atol):
    torch.manual_seed(0)

    arrangement = functools.partial(
        addmm.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64
    )
    application = addmm.application
    tensors = tuple(
        Tensor(ndim, dtype=ninetoothed_dtype) for ndim in (2, 2, 2, 0, 0, 2)
    )
    caller = device
    kernel_name = "addmm"
    output_dir = ninetoothed.generation.CACHE_DIR

    launch_func = _generate_launch_func(
        arrangement,
        application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    input = torch.randn((m, n), dtype=dtype, device=device)
    mat1 = torch.randn((m, k), dtype=dtype, device=device)
    mat2 = torch.randn((k, n), dtype=dtype, device=device)
    beta = torch.randn((), dtype=dtype)
    alpha = torch.randn((), dtype=dtype)
    output = torch.empty(
        (mat1.shape[0], mat2.shape[1]), dtype=mat1.dtype, device=mat1.device
    )

    _run_launch_func(launch_func, input, mat1, mat2, beta, alpha, output)

    expected = torch.addmm(input, mat1, mat2, beta=beta, alpha=alpha)

    assert torch.allclose(output, expected, atol=atol)


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize(
    "dtype, ninetoothed_dtype, atol", ((torch.float16, ninetoothed.float16, 0.01),)
)
@pytest.mark.parametrize("emb_dim", (64,))
@pytest.mark.parametrize("seq_len", (1024,))
@pytest.mark.parametrize("num_heads", (4,))
@pytest.mark.parametrize("batch_size", (2,))
def test_attention(
    batch_size, num_heads, seq_len, emb_dim, dtype, device, ninetoothed_dtype, atol
):
    torch.manual_seed(0)

    arrangement = functools.partial(
        attention.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64
    )
    application = attention.application
    query_, key_, value_, output_ = tuple(
        Tensor(4, dtype=ninetoothed_dtype) for _ in range(4)
    )
    for tensor in (query_, key_, value_, output_):
        tensor.shape = tensor.shape[:-1] + (emb_dim,)
    is_causal_ = Tensor(0, constexpr=True, value=1)
    tensors = (query_, key_, value_, is_causal_, output_)
    caller = device
    kernel_name = "attention"
    output_dir = ninetoothed.generation.CACHE_DIR

    launch_func = _generate_launch_func(
        arrangement,
        application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    shape = (batch_size, num_heads, seq_len, emb_dim)

    query = torch.randn(shape, dtype=dtype, device=device)
    key = torch.randn(shape, dtype=dtype, device=device)
    value = torch.randn(shape, dtype=dtype, device=device)
    is_causal = torch.tensor(True)
    output = torch.empty(shape, dtype=dtype, device=device)

    _run_launch_func(launch_func, query, key, value, is_causal, output)

    expected = F.scaled_dot_product_attention(
        query, key, value, is_causal=True, scale=1
    )

    assert torch.allclose(output, expected, atol=atol)


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize(
    "dtype, ninetoothed_dtype", ((torch.float16, ninetoothed.float16),)
)
@pytest.mark.parametrize("k", (512,))
@pytest.mark.parametrize("n", (512,))
@pytest.mark.parametrize("m", (512,))
def test_matmul(m, n, k, dtype, device, ninetoothed_dtype):
    torch.manual_seed(0)

    arrangement = functools.partial(
        matmul.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64
    )
    application = matmul.application
    tensors = tuple(Tensor(2, dtype=ninetoothed_dtype) for _ in range(3))
    caller = device
    kernel_name = "matmul"
    output_dir = ninetoothed.generation.CACHE_DIR

    launch_func = _generate_launch_func(
        arrangement,
        application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    lhs = torch.randn((m, k), dtype=dtype, device=device)
    rhs = torch.randn((k, n), dtype=dtype, device=device)
    output = torch.empty((lhs.shape[0], rhs.shape[1]), dtype=dtype, device=device)

    _run_launch_func(launch_func, lhs, rhs, output)

    expected = torch.matmul(lhs, rhs)

    assert torch.allclose(output, expected)


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize(
    "dtype, ninetoothed_dtype, rtol, atol",
    ((torch.float16, ninetoothed.float16, 0.001, 0.001),),
)
@pytest.mark.parametrize("s", (3,))
@pytest.mark.parametrize("r", (3,))
@pytest.mark.parametrize("k", (512,))
@pytest.mark.parametrize("w", (16,))
@pytest.mark.parametrize("h", (16,))
@pytest.mark.parametrize("c", (64,))
@pytest.mark.parametrize("n", (4,))
def test_conv2d(n, c, h, w, k, r, s, dtype, device, ninetoothed_dtype, rtol, atol):
    torch.manual_seed(0)

    arrangement = functools.partial(
        conv2d.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64
    )
    application = matmul.application
    tensors = tuple(Tensor(4, dtype=ninetoothed_dtype) for _ in range(3))
    caller = device
    kernel_name = "conv2d"
    output_dir = ninetoothed.generation.CACHE_DIR

    launch_func = _generate_launch_func(
        arrangement,
        application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    p = h - r + 1
    q = w - s + 1

    input = torch.randn(n, c, h, w, dtype=dtype, device=device)
    filter = torch.randn(k, c, r, s, dtype=dtype, device=device)
    output = torch.empty(n, k, p, q, dtype=dtype, device=device)

    _run_launch_func(launch_func, input, filter, output)

    expected = F.conv2d(input, filter)

    assert torch.allclose(output, expected, rtol=rtol, atol=atol)


class _ArgumentTensor(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_void_p),
        ("shape", ctypes.POINTER(ctypes.c_uint64)),
        ("strides", ctypes.POINTER(ctypes.c_int64)),
    ]

    @staticmethod
    def from_torch_tensor(tensor):
        data = ctypes.c_void_p(tensor.data_ptr())
        shape = (ctypes.c_uint64 * len(tensor.shape))(*tensor.shape)
        strides = (ctypes.c_int64 * len(tensor.stride()))(*tensor.stride())

        return _ArgumentTensor(data, shape, strides)


def _run_launch_func(launch_func, *tensors):
    stream = torch.cuda.Stream()

    arg_tensors = tuple(_ArgumentTensor.from_torch_tensor(tensor) for tensor in tensors)

    with torch.cuda.stream(stream):
        launch_func(ctypes.c_void_p(stream.cuda_stream), *arg_tensors)

    stream.synchronize()


def _generate_launch_func(
    arrangement, application, tensors, caller, kernel_name, output_dir
):
    ninetoothed.make(
        arrangement,
        application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    _compile_library(kernel_name, output_dir)
    library = _load_library(kernel_name, output_dir)
    launch_func_name = f"launch_{kernel_name}"
    launch_func = getattr(library, launch_func_name)
    launch_func.argtypes = (ctypes.c_void_p,) + tuple(_ArgumentTensor for _ in tensors)
    launch_func.restype = ctypes.c_int

    return launch_func


def _compile_library(kernel_name, output_dir):
    command = [
        "nvcc",
        "-shared",
        "-Xcompiler",
        "-fPIC",
        "-lcuda",
        "-o",
        output_dir / f"{kernel_name}.so",
        output_dir / f"{kernel_name}.c",
    ]

    subprocess.run(command, check=True)


def _load_library(kernel_name, kernel_dir):
    return ctypes.CDLL(kernel_dir / f"{kernel_name}.so")



import pytest
import torch

from ninetoothed import Tensor, block_size, make
from tests.utils import get_available_devices

BLOCK_SIZE = block_size()


def arrangement(x, BLOCK_SIZE=BLOCK_SIZE):
    return (x.expand((BLOCK_SIZE,)).tile((BLOCK_SIZE,)),)


def application(x):
    x


@pytest.mark.parametrize("device", get_available_devices())
def test_expand(device):
    x = torch.empty((0,), device=device)

    kernel = make(arrangement, application, (Tensor(1),))

    kernel(x)



import random

import pytest
import torch

import ninetoothed
from ninetoothed import Symbol, Tensor
from tests.utils import get_available_devices

BLOCK_SIZE = Symbol("block_size", constexpr=True)


def arrangement(input, output, block_size=BLOCK_SIZE):
    tile_shape = (1,) + tuple(block_size for _ in range(1, input.ndim))

    return input.tile(tile_shape), output.tile(tile_shape)


def application(input, output):
    output = input  # noqa: F841


def to_padded_tensor(input, padding, jagged_dim, block_size=32):
    max_seq_len = input.offsets().diff().max().item()
    output_shape = tuple(
        size if dim != jagged_dim else max_seq_len
        for dim, size in enumerate(input.shape)
    )

    output = torch.empty(output_shape, dtype=input.dtype, device=input.device)

    tensors = (
        Tensor(input.ndim, jagged_dim=jagged_dim, other=padding),
        Tensor(output.ndim),
    )

    kernel = ninetoothed.make(arrangement, application, tensors)

    kernel(input, output, block_size=block_size)

    return output


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("padding", (-1,))
@pytest.mark.parametrize("num_batches", (2, 3, 7, 16))
@pytest.mark.parametrize("jagged_dim", (1, 2))
@pytest.mark.parametrize("ndim", (3,))
def test_to_padded_tensor(ndim, jagged_dim, num_batches, padding, device):
    random.seed(0)
    torch.manual_seed(0)

    def _random_size(lower_bound=1, upper_bound=1024):
        return random.randint(lower_bound, upper_bound)

    def _random_batch_shape(batch_ndim):
        return tuple(_random_size() for _ in range(batch_ndim))

    batch_shape = _random_batch_shape(ndim - 1)

    batches = tuple(
        torch.randn(
            batch_shape[: jagged_dim - 1]
            + (_random_size(),)
            + batch_shape[jagged_dim:],
            device=device,
        )
        for _ in range(num_batches)
    )

    input = torch.nested.nested_tensor(batches, layout=torch.jagged)

    output = to_padded_tensor(input, padding=padding, jagged_dim=jagged_dim)

    expected = torch.nested.to_padded_tensor(input, padding)

    assert output.shape == expected.shape and torch.allclose(output, expected)



import pytest
import torch

import ninetoothed
from ninetoothed import Symbol, Tensor
from tests.utils import get_available_devices


def add(lhs, rhs):
    BLOCK_SIZE = Symbol("BLOCK_SIZE", meta=True)

    @ninetoothed.jit
    def add_kernel(
        lhs: Tensor(1).tile((BLOCK_SIZE,)),
        rhs: Tensor(1).tile((BLOCK_SIZE,)),
        output: Tensor(1).tile((BLOCK_SIZE,)),
    ):
        output = lhs + rhs  # noqa: F841

    output = torch.empty_like(lhs)

    add_kernel(lhs, rhs, output)

    return output


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("dtype", (torch.float32,))
@pytest.mark.parametrize("size", (98432,))
def test(size, dtype, device):
    torch.manual_seed(0)

    input = torch.rand(size, dtype=dtype, device=device)
    other = torch.rand(size, dtype=dtype, device=device)

    output = add(input, other)
    expected = input + other

    assert torch.allclose(output, expected)



import pytest
import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Symbol, Tensor
from tests.utils import get_available_devices


def softmax(input):
    BLOCK_SIZE = Symbol("BLOCK_SIZE", constexpr=True)

    @ninetoothed.jit
    def softmax_kernel(
        input_row: Tensor(2, other=float("-inf")).tile((1, BLOCK_SIZE)),
        output_row: Tensor(2).tile((1, BLOCK_SIZE)),
    ):
        row_minus_max = input_row - ntl.max(input_row)
        numerator = ntl.exp(row_minus_max)
        denominator = ntl.sum(numerator)
        output_row = numerator / denominator  # noqa: F841

    output = torch.empty_like(input)

    softmax_kernel(input, output, BLOCK_SIZE=input.shape[-1])

    return output


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("dtype", (torch.float32,))
@pytest.mark.parametrize("n", (781,))
@pytest.mark.parametrize("m", (1823,))
def test(m, n, dtype, device):
    torch.manual_seed(0)

    input = torch.rand((m, n), dtype=dtype, device=device)

    output = softmax(input)
    expected = torch.softmax(input, dim=-1)

    assert torch.allclose(output, expected)



import random

import pytest
import torch

import ninetoothed
import ninetoothed.language as ntl
import tests.test_matmul as matmul
from ninetoothed import Tensor
from tests.utils import get_available_devices


def arrangement(
    input,
    mat1,
    mat2,
    beta,
    alpha,
    output,
    BLOCK_SIZE_M=matmul.BLOCK_SIZE_M,
    BLOCK_SIZE_N=matmul.BLOCK_SIZE_N,
    BLOCK_SIZE_K=matmul.BLOCK_SIZE_K,
):
    _, _, input_arranged = matmul.arrangement(
        mat1, mat2, input, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K
    )

    mat1_arranged, mat2_arranged, output_arranged = matmul.arrangement(
        mat1, mat2, output, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K
    )

    return input_arranged, mat1_arranged, mat2_arranged, beta, alpha, output_arranged


def application(input, mat1, mat2, beta, alpha, output):
    matmul_output = ntl.zeros(output.shape, dtype=ntl.float32)
    matmul.application(mat1, mat2, matmul_output)
    output = beta * input + alpha * matmul_output


def addmm(input, mat1, mat2, beta=1, alpha=1):
    output = torch.empty(
        (mat1.shape[0], mat2.shape[1]), device=mat1.device, dtype=torch.float16
    )

    addmm_kernel = ninetoothed.make(
        arrangement,
        application,
        (Tensor(2), Tensor(2), Tensor(2), Tensor(0), Tensor(0), Tensor(2)),
    )

    addmm_kernel(input, mat1, mat2, beta, alpha, output)

    return output


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize(
    "dtype, atol", ((torch.float16, 0.075),) + matmul._FLOAT8_E5M2_CONFIG
)
@pytest.mark.parametrize("k", (512,))
@pytest.mark.parametrize("n", (512,))
@pytest.mark.parametrize("m", (512,))
def test(m, n, k, dtype, device, atol):
    random.seed(0)
    torch.manual_seed(0)

    randn_dtype = dtype if dtype != torch.float8_e5m2 else torch.float16

    input = torch.randn((m, n), dtype=randn_dtype, device=device)
    mat1 = torch.randn((m, k), dtype=randn_dtype, device=device)
    mat2 = torch.randn((k, n), dtype=randn_dtype, device=device)
    beta = random.uniform(0, 1)
    alpha = random.uniform(0, 1)

    if dtype == torch.float8_e5m2:
        input = input.to(dtype)
        mat1 = mat1.to(dtype)
        mat2 = mat2.T.to(dtype)

        output = addmm(input, mat1, mat2, beta=beta, alpha=alpha)
        expected = torch.addmm(
            input.to(torch.float16),
            mat1.to(torch.float16),
            mat2.to(torch.float16),
            beta=beta,
            alpha=alpha,
        )
    else:
        output = addmm(input, mat1, mat2, beta=beta, alpha=alpha)
        expected = torch.addmm(input, mat1, mat2, beta=beta, alpha=alpha)

    assert torch.allclose(output, expected, atol=atol)



import functools

import pytest
import torch

import ninetoothed
import tests.test_matmul as matmul
from ninetoothed import Tensor
from tests.utils import get_available_devices


@pytest.mark.parametrize("_device", get_available_devices())
@pytest.mark.parametrize("num_stages", (None, 3, (1, 3)))
@pytest.mark.parametrize("num_warps", (None, 4, (4, 8)))
@pytest.mark.parametrize("block_size_k", (ninetoothed.block_size(), 64))
@pytest.mark.parametrize("block_size_n", (ninetoothed.block_size(), 64))
@pytest.mark.parametrize("block_size_m", (ninetoothed.block_size(), 64))
def test_auto_tuning_generation(
    block_size_m, block_size_n, block_size_k, num_warps, num_stages, _device
):
    auto_tuning_should_be_disabled = (
        isinstance(block_size_m, int)
        and isinstance(block_size_n, int)
        and isinstance(block_size_k, int)
        and not isinstance(num_warps, tuple)
        and not isinstance(num_stages, tuple)
    )

    arrangement = functools.partial(
        matmul.arrangement,
        BLOCK_SIZE_M=block_size_m,
        BLOCK_SIZE_N=block_size_n,
        BLOCK_SIZE_K=block_size_k,
    )

    application = matmul.application

    tensors = (Tensor(2), Tensor(2), Tensor(2))

    kernel = ninetoothed.make(
        arrangement, application, tensors, num_warps=num_warps, num_stages=num_stages
    )

    source_file = kernel._source

    with open(source_file) as f:
        contents = f.read()

        if auto_tuning_should_be_disabled:
            assert "application_with_auto_tuning" not in contents
            assert "num_warps=" in contents and "num_stages=" in contents
        else:
            assert "application_with_auto_tuning" in contents


@pytest.mark.parametrize("_device", get_available_devices())
def test_arrangement_returning_a_single_tensor(_device):
    def arrangement(x):
        return x.tile((ninetoothed.block_size(),))

    def application(x):
        x

    ninetoothed.make(arrangement, application, (Tensor(1),))


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("num_indices", (10,))
@pytest.mark.parametrize("num_cols", (128,))
@pytest.mark.parametrize("num_rows", (1024,))
def test_squeezing_the_innermost_level(num_rows, num_cols, num_indices, device):
    torch.manual_seed(0)

    def arrangement(input, indices, output):
        input_arranged = input.tile((1, -1)).squeeze(1)
        input_arranged.dtype = input_arranged.dtype.squeeze(0)

        indices_arranged = indices.tile((1,))

        output_arranged = (
            output.tile((1, -1))
            .tile((-1, -1))
            .squeeze(0)
            .expand((input_arranged.shape[0],))
        )
        output_arranged.dtype = output_arranged.dtype.squeeze(1)
        output_arranged.dtype.dtype = output_arranged.dtype.dtype.squeeze(0)

        return input_arranged, indices_arranged, output_arranged

    def application(input, index, output):
        row = index

        output[row] = input

    shape_options = {"constexpr": True}

    tensors = (
        Tensor(2, shape_options=shape_options),
        Tensor(1, shape_options=shape_options),
        Tensor(2, shape_options=shape_options),
    )

    kernel = ninetoothed.make(arrangement, application, tensors)

    input = torch.randn((num_indices, num_cols), device=device)
    indices = torch.randint(0, num_rows, (num_indices,), device=device)
    output = torch.empty((num_rows, num_cols), device=device)
    expected = output.clone()

    kernel(input, indices, output)

    for i in range(num_indices):
        expected[indices[i], :] = input[i, :]

    assert output.shape == expected.shape and torch.allclose(output, expected)


@pytest.mark.parametrize("device", get_available_devices())
def test_unsqueezing_the_outermost_level(device):
    def arrangement(x):
        return x.tile((ninetoothed.block_size(),)).unsqueeze(0)

    def application(x):
        x

    kernel = ninetoothed.make(arrangement, application, (Tensor(1),))

    kernel(torch.randn((0,), device=device))



import numpy as np

import ninetoothed
from ninetoothed import Tensor


def test_unsqueeze():
    x = Tensor(1)

    block_size_m = ninetoothed.block_size()
    block_size_n = ninetoothed.block_size()

    subs = {x: {"shape": (5,)}, block_size_m: 2, block_size_n: 3}

    x_unsqueezed = x.unsqueeze(0)

    assert np.array_equal(x_unsqueezed.eval(subs), np.array([[0, 1, 2, 3, 4]]))

    x_unsqueezed_expanded = x_unsqueezed.expand((3, -1))

    assert np.array_equal(
        x_unsqueezed_expanded.eval(subs),
        np.array(
            [
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
            ]
        ),
    )

    x_unsqueezed_expanded_tiled = x_unsqueezed_expanded.tile(
        (block_size_m, block_size_n)
    )

    assert np.array_equal(
        x_unsqueezed_expanded_tiled.eval(subs),
        np.array(
            [
                [[[0, 1, 2], [0, 1, 2]], [[3, 4, -1], [3, 4, -1]]],
                [[[0, 1, 2], [-1, -1, -1]], [[3, 4, -1], [-1, -1, -1]]],
            ]
        ),
    )

    x_tiled_unsqueezed = x.tile((block_size_n,)).unsqueeze(0)

    assert np.array_equal(
        x_tiled_unsqueezed.eval(subs), np.array([[[0, 1, 2], [3, 4, -1]]])
    )

    x_tiled_unsqueezed_expanded = x_tiled_unsqueezed.expand((3, -1))

    assert np.array_equal(
        x_tiled_unsqueezed_expanded.eval(subs),
        np.array(
            [[[0, 1, 2], [3, 4, -1]], [[0, 1, 2], [3, 4, -1]], [[0, 1, 2], [3, 4, -1]]]
        ),
    )



import pytest
import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Symbol, Tensor
from tests.utils import get_available_devices

BLOCK_SIZE_M = Symbol("BLOCK_SIZE_M", meta=True)
BLOCK_SIZE_N = Symbol("BLOCK_SIZE_N", meta=True)
BLOCK_SIZE_K = Symbol("BLOCK_SIZE_K", meta=True)


def arrangement(
    lhs,
    rhs,
    output,
    BLOCK_SIZE_M=BLOCK_SIZE_M,
    BLOCK_SIZE_N=BLOCK_SIZE_N,
    BLOCK_SIZE_K=BLOCK_SIZE_K,
):
    output_tiled = output.tile((BLOCK_SIZE_M, BLOCK_SIZE_N))

    lhs_tiled = (
        lhs.tile((BLOCK_SIZE_M, BLOCK_SIZE_K))
        .tile((1, -1))
        .expand((-1, output_tiled.shape[1]))
    )
    lhs_tiled.dtype = lhs_tiled.dtype.squeeze(0)

    rhs_tiled = (
        rhs.tile((BLOCK_SIZE_K, BLOCK_SIZE_N))
        .tile((-1, 1))
        .expand((output_tiled.shape[0], -1))
    )
    rhs_tiled.dtype = rhs_tiled.dtype.squeeze(1)

    return lhs_tiled, rhs_tiled, output_tiled


def application(lhs, rhs, output):
    accumulator = ntl.zeros(output.shape, dtype=ntl.float32)
    for k in range(lhs.shape[0]):
        accumulator += ntl.dot(lhs[k], rhs[k])
    output = accumulator.to(ntl.float16)


def matmul(lhs, rhs):
    output = torch.empty(
        (lhs.shape[0], rhs.shape[1]), device=lhs.device, dtype=torch.float16
    )

    matmul_kernel = ninetoothed.make(
        arrangement, application, (Tensor(2), Tensor(2), Tensor(2))
    )

    matmul_kernel(lhs, rhs, output)

    return output


_FLOAT8_E5M2_CONFIG = (
    ((torch.float8_e5m2, 0.125),) if hasattr(torch, "float8_e5m2") else ()
)


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("dtype, atol", ((torch.float16, 1e-8),) + _FLOAT8_E5M2_CONFIG)
@pytest.mark.parametrize("k", (512,))
@pytest.mark.parametrize("n", (512,))
@pytest.mark.parametrize("m", (512,))
def test(m, n, k, dtype, device, atol):
    torch.manual_seed(0)

    randn_dtype = dtype if dtype != torch.float8_e5m2 else torch.float16

    input = torch.randn((m, k), dtype=randn_dtype, device=device)
    other = torch.randn((k, n), dtype=randn_dtype, device=device)

    if dtype == torch.float8_e5m2:
        input = input.to(dtype)
        other = other.T.to(dtype)

        output = matmul(input, other)
        expected = torch.matmul(input.to(torch.float16), other.to(torch.float16))
    else:
        output = matmul(input, other)
        expected = torch.matmul(input, other)

    assert torch.allclose(output, expected, atol=atol)



import random

import pytest
import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Tensor, block_size
from tests.utils import get_available_devices


def arrangement(input, p, seed, output, BLOCK_SIZE=block_size()):
    return input.tile((BLOCK_SIZE,)), p, seed, output.tile((BLOCK_SIZE,))


def application(input, p, seed, output):
    output = ntl.where(ntl.rand(seed, input.offsets()) > p, input / (1 - p), 0)  # noqa: F841


def dropout(input, p=0.5):
    seed = random.randrange(0, 2**31)
    output = torch.empty_like(input)

    tensors = (Tensor(1), Tensor(0), Tensor(0), Tensor(1))
    dropout_kernel = ninetoothed.make(arrangement, application, tensors)

    dropout_kernel(input, p, seed, output)

    return output


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("dtype", (torch.float16,))
@pytest.mark.parametrize("size", (349,))
def test(size, dtype, device):
    random.seed(0)
    torch.manual_seed(0)

    input = torch.randn(size, dtype=dtype, device=device)
    p = 0.3

    output = dropout(input, p=p)

    assert input.shape == output.shape

    non_zero_ratio = output.nonzero().numel() / input.numel()

    assert abs(non_zero_ratio - (1 - p)) < 0.05

    assert torch.allclose(output[output != 0], input[output != 0] / (1 - p))



import pytest
import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Tensor
from tests.utils import get_available_devices


def arrangement(input, output, BLOCK_SIZE=1024):
    input_arranged = input.tile((BLOCK_SIZE,))
    output_arranged = output.tile((-1,)).expand(input_arranged.shape)

    return input_arranged, output_arranged


def application(input, output):
    ntl.atomic_add(output.data_ptr(), ntl.sum(input))


def sum(input):
    output = torch.zeros(1, dtype=input.dtype, device=input.device)

    tensors = (Tensor(1), Tensor(1, shape_options={"constexpr": True}))

    sum_kernel = ninetoothed.make(arrangement, application, tensors)

    sum_kernel(input, output)

    return output[0]


@pytest.mark.parametrize("device", get_available_devices())
@pytest.mark.parametrize("size", (64180,))
def test_data_ptr(size, device):
    torch.manual_seed(0)

    input = torch.randn(size, device=device)

    output = sum(input)

    expected = torch.sum(input)

    assert output.shape == expected.shape and torch.allclose(output, expected)
